from unittest.mock import Mock

from ..tasks import circulars, gracedb


def test_create_circular(monkeypatch):
    """Test that the compose circulars method is called with the correct
    input parameters."""
    superevent_id = 'S1234'
    mock_compose = Mock()
    monkeypatch.setattr('ligo.followup_advocate.compose', mock_compose)

    # call create_circular
    circulars.create_circular(superevent_id)
    mock_compose.assert_called_once_with('S1234', client=gracedb.client)



"""Source Parameter Estimation with LALInference."""
from distutils.spawn import find_executable
from distutils.dir_util import mkpath
import glob
import json
import math
import os
import shutil
import subprocess
import tempfile

from celery import group
from glue.lal import Cache
from gwdatafind import find_urls
from gwpy.timeseries import StateVector
import lal
import lalsimulation

from .. import app
from ..jinja import env
from . import condor
from . import gracedb


ini_name = 'online_pe.ini'

executables = {'datafind': 'gw_data_find',
               'mergeNSscript': 'lalinference_nest2pos',
               'mergeMCMCscript': 'cbcBayesMCMC2pos',
               'combinePTMCMCh5script': 'cbcBayesCombinePTMCMCh5s',
               'resultspage': 'cbcBayesPostProc',
               'segfind': 'ligolw_segment_query',
               'ligolw_print': 'ligolw_print',
               'coherencetest': 'lalinference_coherence_test',
               'lalinferencenest': 'lalinference_nest',
               'lalinferencemcmc': 'lalinference_mcmc',
               'lalinferencebambi': 'lalinference_bambi',
               'lalinferencedatadump': 'lalinference_datadump',
               'ligo-skymap-from-samples': 'ligo-skymap-from-samples',
               'ligo-skymap-plot': 'ligo-skymap-plot',
               'processareas': 'process_areas',
               'computeroqweights': 'lalinference_compute_roq_weights',
               'mpiwrapper': 'lalinference_mpi_wrapper',
               'gracedb': 'gracedb',
               'ppanalysis': 'cbcBayesPPAnalysis',
               'pos_to_sim_inspiral': 'cbcBayesPosToSimInspiral'}

flow = 20.0

# number of samples between end time of the data used for psd estimation and
# start time of data for PE
padding = 16

# default number of realizations used for PSD estimation
default_num_of_realizations = 32


def _webdir(graceid):
    """Return webdir filled in .ini file"""
    return os.getenv('HOME') + '/public_html/online_pe/' + graceid


def _ifos(event):
    """Return ifos"""
    return event['extra_attributes']['CoincInspiral']['ifos'].split(',')


def _chirplen(singleinspiral):
    """Return chirplen"""
    return lalsimulation.SimInspiralChirpTimeBound(
               flow,
               singleinspiral['mass1'] * lal.MSUN_SI,
               singleinspiral['mass2'] * lal.MSUN_SI,
               0.0, 0.0
           )


def _round_up_to_power_of_two(x):
    """Return smallest power of two exceeding x"""
    return 2**math.ceil(math.log(x, 2))


def _seglen(singleinspiraltable):
    """Return seglen"""
    return max([_round_up_to_power_of_two(max(4.0, _chirplen(sngl) + 2.0))
                for sngl in singleinspiraltable])


def _freq_dict(freq, ifos):
    """Return dictionary whose keys are ifos and items are frequencies"""
    return dict((ifo, freq) for ifo in ifos)


def _fstop(singleinspiral):
    """Return final frequency"""
    return lalsimulation.IMRPhenomDGetPeakFreq(
               singleinspiral['mass1'], singleinspiral['mass2'], 0.0, 0.0
           )


def _srate(singleinspiraltable):
    """Return srate we should use"""
    return _round_up_to_power_of_two(
               max([_fstop(sngl) for sngl in singleinspiraltable])
           ) * 2


def _start_end(trigtime, seglen, num_of_realizations):
    """Return gps start time and end time

    Parameters
    ----------
    trigtime : float
        The time of target trigger
    seglen : int
        The length of the segment used to calculate overlap between data and
        waveform
    num_of_realizations : int
        The number of noise realizations used for PSD estimation

    Return
    ------
    start : float
        GPS start time of the whole data used for Parameter Estimation
    end : float
        GPS end time of the whole data used for Parameter Estimation
    """
    end = trigtime + 2
    return end - seglen - padding - num_of_realizations * seglen, end


def _start_end_of_science_segment_for_one_ifo(start, end, ifo, frametype):
    """Return gps start time and end time of correctly calibrated and
    observing-intent data on the time when interferometers are locked. First
    this function searches for available data. If parts of the data are
    unavailabile, this function throws away data except for the last
    continuously available data. Next, this function cheks the Bit 0 (HOFT_OK),
    1 (OBSERVATION_INTENT) and 2 (OBSERVATION_READY) in GDS-CALIB_STATE_VECTOR
    and extracts the last continous data which have 1 in all the three bits.
    Finally this function returns gps start time and end time of the resultant
    data. This function looks for data only for one specified ifo.

    Parameters
    ----------
    start : float
        GPS start time of the time window over which data is searched for.
    end : float
        GPS end time of the time window over which data is searched for.
    ifo : str
    frametype : str

    Return
    ------
    start : float
        GPS start time of found science segment
    end : float
        GPS end time of found science segment
    """
    # get gps start and end time of available data
    datacache = Cache.from_urls(find_urls(ifo[0], frametype, start, end))
    available_segment = datacache.to_segmentlistdict()[ifo[0]][-1]
    start, end = available_segment[0], available_segment[1]

    # check whether data is calibrated correctly, observing-intent and taken
    # while the inferferometers are locked
    flag = StateVector.read(
               datacache, app.conf['state_vector_channel_names'][ifo],
               start=start, end=end,
               bits=["HOFT_OK", "OBSERVATION_INTENT", "OBSERVATION_READY"]
           ).to_dqflags()
    pe_segment = (flag['HOFT_OK'].active - ~flag['OBSERVATION_INTENT'].active -
                  ~flag['OBSERVATION_READY'].active)[-1]
    return pe_segment[0], pe_segment[1]


def _start_end_of_science_segment(trigtime, seglen, ifos, frametype_dict):
    """Calculate gps start time and end time of ready-for-PE data with
    _start_end_of_science_segment for each ifo and return maximum of start time
    and minimum of end time"""
    start, end = _start_end(trigtime, seglen, default_num_of_realizations)
    for ifo in ifos:
        start, end = _start_end_of_science_segment_for_one_ifo(
                         start, end, ifo, frametype_dict[ifo]
                     )
    return start, end


def _psdstart_psdlength(start, trigtime, seglen):
    """Return gps start time and length of data for PSD estimation as a list.
    Note that psdlength has to be seglen multiplied by an integer"""
    psdlength = \
        math.floor(trigtime + 2 - seglen - padding - start) // seglen * seglen
    return [trigtime + 2 - seglen - padding - psdlength, psdlength]


def _find_appropriate_frametype_psdstart_psdlength(trigtime, seglen,
                                                   ifos, superevent_id=None):
    """Return appropriate frametype, psdstart and psdlength. This function sets
    long enough psdlength first and shorten psdlength depending on whether data
    is available, correctly calibrated, observing-intent and taken while the
    interferometers are locked. This function first searches for low-latency
    frame data and, if they are not available, searches for high-latency frame
    data. If enough data is not found finally, raise exception and report
    failure to GraceDB"""
    # First search for low-latecy frame data
    frametype_dict = app.conf['low_latency_frame_types']
    start, end = _start_end_of_science_segment(
                     trigtime, seglen, ifos, frametype_dict
                 )
    start_threshold, end_threshold = _start_end(trigtime, seglen, 1)
    # Here we take into account the case where very recent data is not
    # available at this time but it will be available at the actual start time
    # of Parameter Estimation
    if start <= start_threshold:
        return frametype_dict, _psdstart_psdlength(start, trigtime, seglen)

    # If part of low-latency data has already vanished, search for high-latency
    # frame data
    frametype_dict = app.conf['high_latency_frame_types']
    start, end = _start_end_of_science_segment(
                     trigtime, seglen, ifos, frametype_dict
                 )
    if start <= start_threshold and end >= end_threshold:
        return frametype_dict, _psdstart_psdlength(start, trigtime, seglen)
    else:
        if superevent_id is not None:
            gracedb.upload.delay(
                filecontents=None, filename=None,
                graceid=superevent_id,
                message='Available data is not long enough, and ' +
                        'Parameter Estimation will never start automatically.',
                tags='pe'
            )
        raise


@app.task(shared=False)
def prepare_ini(event, superevent_id=None):
    """Determine an appropriate PE settings for the target event and return ini
    file content
    """
    # Get template of .ini file
    ini_template = env.get_template(ini_name)

    # Download event's info to determine PE settings
    singleinspiraltable = event['extra_attributes']['SingleInspiral']
    trigtime = event['gpstime']

    # fill out the ini template and return the resultant content
    ifos = _ifos(event)
    seglen = _seglen(singleinspiraltable)
    # FIXME: seglen here might not be actual seglen if ROQ is used
    frametypes, psdstart_psdlength = \
        _find_appropriate_frametype_psdstart_psdlength(
            trigtime, seglen, ifos, superevent_id
        )
    ini_settings = {
        'service_url': gracedb.client._service_url,
        'types': frametypes,
        'channels': app.conf['strain_channel_names'],
        'webdir': _webdir(event['graceid']),
        'paths': [{'name': name, 'path': find_executable(executable)}
                  for name, executable in executables.items()],
        'q': min([sngl['mass2'] / sngl['mass1']
                  for sngl in singleinspiraltable]),
        'ifos': ifos,
        'seglen': seglen,
        'flow': _freq_dict(flow, ifos),
        'srate': _srate(singleinspiraltable),
        'psd-start-time': psdstart_psdlength[0],
        'psd-length': psdstart_psdlength[1]
    }
    return ini_template.render(ini_settings)


@app.task(shared=False)
def dag_prepare(rundir, ini_contents, preferred_event_id, superevent_id):
    """Create a Condor DAG to run LALInference on a given event.

    Parameters
    ----------
    rundir : str
        The path to a run directory where the DAG file exits
    ini_contents : str
        The content of online_pe.ini
    preferred_event_id : str
        The GraceDb ID of a target preferred event
    superevent_id : str
        The GraceDb ID of a target superevent

    Returns
    -------
    submit_file : str
        The path to the .sub file
    """
    # write down .ini file in the run directory
    path_to_ini = rundir + '/' + ini_name
    with open(path_to_ini, 'w') as f:
        f.write(ini_contents)

    # run lalinference_pipe
    gracedb.upload.delay(
        filecontents=None, filename=None, graceid=superevent_id,
        message='starting LALInference online parameter estimation',
        tags='pe'
    )
    try:
        subprocess.run(['lalinference_pipe', '--run-path', rundir,
                        '--gid', preferred_event_id, path_to_ini],
                       stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                       check=True)
        subprocess.run(['condor_submit_dag', '-no_submit',
                        rundir + '/multidag.dag'],
                       stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                       check=True)
    except subprocess.CalledProcessError as e:
        contents = b'args:\n' + json.dumps(e.args[1]).encode('utf-8') + \
                   b'\n\nstdout:\n' + e.stdout + b'\n\nstderr:\n' + e.stderr
        gracedb.upload.delay(
            filecontents=contents, filename='pe_dag.log',
            graceid=superevent_id,
            message='Failed to prepare DAG', tags='pe'
        )
        shutil.rmtree(rundir)
        raise

    return rundir + '/multidag.dag.condor.sub'


@app.task(ignore_result=True, shared=False)
def job_error_notification(request, exc, traceback, superevent_id):
    """Upload notification when condor.submit terminates unexpectedly.

    Parameters
    ----------
    request : Context (placeholder)
        Task request variables
    exc : Exception
        Exception rased by condor.submit
    traceback : str (placeholder)
        Traceback message from a task
    superevent_id : str
        The GraceDb ID of a target superevent
    """
    if type(exc) is condor.JobAborted:
        gracedb.upload.delay(
            filecontents=None, filename=None, graceid=superevent_id,
            message='Job was aborted.', tags='pe'
        )
    elif type(exc) is condor.JobFailed:
        gracedb.upload.delay(
            filecontents=None, filename=None, graceid=superevent_id,
            message='Job failed', tags='pe'
        )


@app.task(ignore_result=True, shared=False)
def upload_result(webdir, filename, graceid, message, tag):
    """Upload a PE result

    Parameters
    ----------
    graceid : str
        The GraceDb ID.
    """
    paths = list(glob.iglob(webdir + '/**/' + filename, recursive=True))
    if len(paths) == 1:
        with open(paths[0], 'rb') as f:
            contents = f.read()
        gracedb.upload.delay(
            contents, filename,
            graceid, message, tag
        )


@app.task(ignore_result=True, shared=False)
def clean_up(rundir):
    """Clean up a run directory.

    Parameters
    ----------
    rundir : str
        The path to a run directory where the DAG file exits
    """
    shutil.rmtree(rundir)


def dag_finished(rundir, preferred_event_id, superevent_id):
    """Upload PE results and clean up run directory

    Parameters
    ----------
    rundir : str
        The path to a run directory where the DAG file exits
    preferred_event_id : str
        The GraceDb ID of a target preferred event
    superevent_id : str
        The GraceDb ID of a target superevent

    Returns
    -------
    tasks : canvas
        The work-flow for uploading PE results
    """
    # get webdir where the results are outputted
    webdir = _webdir(preferred_event_id)

    return group(
        gracedb.upload.si(
            filecontents=None, filename=None, graceid=superevent_id,
            message='LALInference online parameter estimation finished.',
            tags='pe'
        ),
        upload_result.si(
            webdir, 'LALInference.fits.gz', superevent_id,
            'LALInference FITS sky map', ['pe', 'sky_loc']
        ),
        upload_result.si(
            webdir, 'extrinsic.png', superevent_id,
            'Corner plot for extrinsic parameters', 'pe'
        ),
        upload_result.si(
            webdir, 'intrinsic.png', superevent_id,
            'Corner plot for intrinsic parameters', 'pe'
        ),
        upload_result.si(
            webdir, 'sourceFrame.png', superevent_id,
            'Corner plot for source frame parameters', 'pe'
        )
    ) | clean_up.si(rundir)


@app.task(ignore_result=True, shared=False)
def start_pe(ini_contents, preferred_event_id, superevent_id):
    """Run LALInference on a given event.

    Parameters
    ----------
    ini_contents : str
        The content of online_pe.ini
    preferred_event_id : str
        The GraceDb ID of a target preferred event
    superevent_id : str
        The GraceDb ID of a target superevent
    """
    # make a run directory
    lalinference_dir = os.path.expanduser('~/.cache/lalinference')
    mkpath(lalinference_dir)
    rundir = tempfile.mkdtemp(dir=lalinference_dir)

    (
        dag_prepare.s(rundir, ini_contents, preferred_event_id, superevent_id)
        |
        condor.submit.s().on_error(job_error_notification.s(superevent_id))
        |
        dag_finished(rundir, preferred_event_id, superevent_id)
    ).delay()



"""Celery application initialization."""

from celery import Celery

from ._version import get_versions
from .conf import playground
from . import sentry

__all__ = ('app',)

__version__ = get_versions()['version']
del get_versions

# Use redis broker, because it supports locks (and thus singleton tasks).
app = Celery(__name__, broker='redis://', autofinalize=False)
"""Celery application object."""

# Register all tasks.
app.autodiscover_tasks([__name__])

# Add default configuration.
app.add_defaults(playground)
app.finalize()

# Customize configuration from environment variable.
app.config_from_envvar('CELERY_CONFIG_MODULE', silent=True)

# Use the same URL for both the result backend and the broker.
app.conf['result_backend'] = app.conf.broker_url

sentry.configure()



"""Module containing the computation of p_astro by source category
   See https://dcc.ligo.org/LIGO-T1800072 for details.
"""
import json

from celery.utils.log import get_task_logger
import numpy as np

from ..import app

log = get_task_logger(__name__)


def p_astro_update(category, event_bayesfac_dict, mean_values_dict):
    """
    This is the function that computes `p_astro` for a new
    event using mean values of Poisson expected counts
    constructed from all the previous events. This is
    the function that will be invoked with every new
    GraceDB entry.

    Parameters
    ----------
    category : string
        source category
    event_bayesfac_dict : dictionary
        event Bayes factors
    mean_values_dict : dictionary
        mean values of Poisson counts

    Returns
    -------
    p_astro : float
        p_astro by source category
    """
    if category == "Terr":
        numerator = mean_values_dict["Terr"]
    else:
        numerator = \
            event_bayesfac_dict[category]*mean_values_dict[category]

    denominator = mean_values_dict["Terr"] + \
        np.sum([mean_values_dict[key]*event_bayesfac_dict[key]
                for key in event_bayesfac_dict.keys()])

    return numerator/denominator


@app.task(shared=False)
def compute_p_astro(snr, far, mass1, mass2):
    """
    Task to compute `p_astro` by source category.

    Parameters
    ----------
    snr : float
        event's snr
    far : float
        event's cfar
    mass1 : float
        event's mass1
    mass2 : float
        event's mass2

    Returns
    -------
    p_astros : str
        JSON dump of the p_astro by source category

    Example
    -------
    >>> p_astros = json.loads(compute_p_astro(files))
    >>> p_astros
    {'BNS': 0.999, 'BBH': 0.0, 'NSBH': 0.0, 'Terr': 0.001}
    """
    mean_bns = 2.07613829518
    mean_nsbh = 1.65747751787
    mean_bbh = 11.8941873831
    mean_terr = 3923
    snr_star = 8.5
    far_star = 1/(30*86400)
    num_bins = 3

    # Fix mean values (1 per source category) from
    mean_values_dict = {"BNS": mean_bns,
                        "NSBH": mean_nsbh,
                        "BBH": mean_bbh,
                        "Terr": mean_terr}

    # These are the bayes factor values that need to be
    # constructed with every new event/GraceDB upload
    fground = 3*snr_star/(snr**4)
    bground = far/far_star

    a_hat_bns = int(mass1 <= 3 and mass2 <= 3)
    a_hat_bbh = int(mass2 > 3 and mass2 > 3)
    a_hat_nsbh = int(min([mass1, mass2]) <= 3 and max([mass1, mass2]) > 3)

    rescaled_fb = num_bins*fground/bground
    bns_bayesfac = a_hat_bns*rescaled_fb
    nsbh_bayesfac = a_hat_nsbh*rescaled_fb
    bbh_bayesfac = a_hat_bbh*rescaled_fb
    event_bayesfac_dict = {"BNS": bns_bayesfac,
                           "NSBH": nsbh_bayesfac,
                           "BBH": bbh_bayesfac}

    # Compute the p-astro values for each source category
    # using the mean values
    p_astro_values = {}
    for category in mean_values_dict:
        p_astro_values[category] = \
            p_astro_update(category=category,
                           event_bayesfac_dict=event_bayesfac_dict,
                           mean_values_dict=mean_values_dict)

    return json.dumps(p_astro_values)



"""GWCelery application configuration.

This module defines configuration variables and default values, including both
:doc:`generic options for Celery <celery:userguide/configuration>` as well as
options that control the behavior of specific GWCelery :mod:`~gwcelery.tasks`.

To override the configuration, define the ``CELERY_CONFIG_MODULE`` environment
variable to the fully qualified name of any Python module that can be located
in :obj:`sys.path`, including any of the following presets:

 * :mod:`gwcelery.conf.development`
 * :mod:`gwcelery.conf.playground` (the default)
 * :mod:`gwcelery.conf.production`
 * :mod:`gwcelery.conf.test`
"""

# Celery application settings.
# Use pickle serializer, because it supports byte values.

accept_content = ['json', 'pickle']
event_serializer = 'json'
result_serializer = 'pickle'
task_serializer = 'pickle'

# GWCelery-specific settings.

lvalert_host = 'lvalert-playground.cgca.uwm.edu'
"""LVAlert host."""

gracedb_host = 'gracedb-playground.ligo.org'
"""GraceDb host."""

gcn_broker_address = ':5341'
"""The VOEvent broker will bind to this address to send GCNs.
This should be a string of the form `host:port`. If `host` is empty,
then listen on all available interfaces."""

gcn_broker_accept_addresses = []
"""List of hosts from which the broker will accept connections.
If empty, then completely disable the broker's broadcast capability."""

gcn_client_address = ''
"""The VOEvent listener will connect to this address to receive GCNs.
If empty, then completely disable the GCN listener."""

superevent_d_t_start = {'gstlal': 1.0,
                        'spiir': 1.0,
                        'pycbc': 1.0,
                        'mbtaonline': 1.0}
"""Pipeline based lower extent of superevent segments.
For cwb and lib this is decided from extra attributes."""

superevent_d_t_end = {'gstlal': 1.0,
                      'spiir': 1.0,
                      'pycbc': 1.0,
                      'mbtaonline': 1.0}
"""Pipeline based upper extent of superevent segments
For cwb and lib this is decided from extra attributes."""

superevent_query_d_t_start = 100.
"""Lower extent of superevents query"""

superevent_query_d_t_end = 100.
"""Upper extent of superevents query"""

superevent_default_d_t_start = 1.0
"""Default lower extent of superevent segments"""

superevent_default_d_t_end = 1.0
"""Default upper extent for superevent segments"""

superevent_far_threshold = 1 / 3600
"""Maximum false alarm rate to consider events superevents."""

preliminary_alert_far_threshold = {'cbc': 1 / (30 * 86400),
                                   'burst': 1 / (365 * 86400),
                                   'test': 1 / (30 * 86400)}
"""Group specific maximum false alarm rate to consider
sending preliminary alerts."""

preliminary_alert_trials_factor = dict(cbc=5.0, burst=5.0)
"""Trials factor corresponding to trigger categories.
For CBC, trials factor is the number of pipelines plus the external
coincidence search. For Burst, this is the total number of searches
plus the external coincidence search.
CBC pipelines are gstlal, pycbc, mbtaonline, spiir.
Burst searches are cwb.allsky, cwb.bbh, cwb.imbh and olib.allsky."""

orchestrator_timeout = 15.0
"""The orchestrator will wait this many seconds from the time of the
creation of a new superevent to the time that annotations begin, in order
to let the superevent manager's decision on the preferred event
stabilize."""

check_vector_prepost = {'gstlal': [2, 2],
                        'spiir': [2, 2],
                        'pycbc': [2, 2],
                        'MBTAOnline': [2, 2],
                        'oLIB': [0.5, 0.5],
                        'LIB': [0.5, 0.5],
                        'CWB': [0.5, 0.5],
                        'HardwareInjection': [2, 2],
                        'Swift': [2, 2],
                        'Fermi': [2, 2],
                        'SNEWS': [2, 2]}
"""Seconds before and after the superevent start and end times which the DQ
vector check will include in its check. Pipeline dependent."""

uses_gatedhoft = {'gstlal': False,
                  'spiir': True,
                  'pycbc': True,
                  'MBTAOnline': True,
                  'oLIB': False,
                  'LIB': False,
                  'CWB': True,
                  'HardwareInjection': False,
                  'Swift': False,
                  'Fermi': False,
                  'SNEWS': False}
"""Whether or not a pipeline uses gated h(t). Determines whether or not
the DMT-DQ_VECTOR will be analyzed for data quality."""

llhoft_glob = '/dev/shm/llhoft/{detector}/*.gwf'
"""File glob for low-latency h(t) frames."""

llhoft_channels = {
    'H1:DMT-DQ_VECTOR': 'dmt_dq_vector_bits',
    'L1:DMT-DQ_VECTOR': 'dmt_dq_vector_bits',
    'H1:GDS-CALIB_STATE_VECTOR': 'state_vector_bits',
    'L1:GDS-CALIB_STATE_VECTOR': 'state_vector_bits',
    'V1:DQ_ANALYSIS_STATE_VECTOR': 'state_vector_bits',
    #  Virgo DQ veto streams, should be implemented before September O2 replay
    #  'V1:DQ_VETO_MBTA': 'no_dq_veto_mbta_bits',
    #  'V1:DQ_VETO_CWB': 'no_dq_veto_cwb_bits',
    #  'V1:DQ_VETO_GSTLAL': 'no_dq_veto_gstlal_bits',
    #  'V1:DQ_VETO_OLIB': 'no_dq_veto_olib_bits',
    #  'V1:DQ_VETO_PYCBC': 'no_dq_veto_pycbc_bits',
                  }
"""Low-latency h(t) state vector configuration. This is a dictionary consisting
of a channel and its bitmask, as defined in :mod:`gwcelery.tasks.detchar`."""

idq_channels = ['H1:IDQ-PGLITCH_RANDOM_FOREST_16_4096',
                'L1:IDQ-PGLITCH_RANDOM_FOREST_16_4096']
"""Low-latency iDQ p(glitch) channel names"""

idq_pglitch_thresh = 0.95
"""Minimum p(glitch) reported by iDQ required before notice is posted to
GraceDb"""

p_astro_gstlal_ln_likelihood_threshold = 6
"""log likelihood threshold"""

p_astro_gstlal_ranking_pdf = '/home/gstlalcbc/observing/3/online/trigs/rankingstat_pdf.xml.gz'  # noqa: E501

low_latency_frame_types = {'H1': 'H1_O2_llhoft',
                           'L1': 'L1_O2_llhoft',
                           'V1': 'V1_O2_llhoft'}
"""Types of frames used in Parameter Estimation with LALInference (see
:mod:`gwcelery.tasks.lalinference`)"""

high_latency_frame_types = {'H1': 'None',
                            'L1': 'None',
                            'V1': 'None'}
"""Types of nonllhoft-frames used in Parameter Estimation with LALInference.
They do not exist for O2Replay data. (see
:mod:`gwcelery.tasks.lalinference`)"""

strain_channel_names = {'H1': 'H1:GDS-CALIB_STRAIN_O2Replay',
                        'L1': 'L1:GDS-CALIB_STRAIN_O2Replay',
                        'V1': 'V1:Hrec_hoft_16384Hz_O2Replay'}
"""Names of h(t) channels used in Parameter Estimation with LALInference (see
:mod:`gwcelery.tasks.lalinference`)"""

state_vector_channel_names = {'H1': 'H1:GDS-CALIB_STATE_VECTOR',
                              'L1': 'L1:GDS-CALIB_STATE_VECTOR',
                              'V1': 'V1:DQ_ANALYSIS_STATE_VECTOR'}
"""Names of state vector channels used in Parameter Estimation with
LALInference (see :mod:`gwcelery.tasks.lalinference`)"""

pe_threshold = 1.0 / (14 * 86400)
"""FAR threshold in Hz for Parameter Estimation. PE group now applies
1/(2 weeks) as a threshold. 86400 seconds = 1 day and 14 days = 2 weeks."""



"""Subsystem for sending, receiving, and processing Gamma-ray Coordinates
Network [GCN]_ notices.

References
----------

.. [GCN] https://gcn.gsfc.nasa.gov
"""
import socket
import subprocess
import sys
import tempfile

from celery.utils.log import get_task_logger
from celery_eternal import EternalTask, EternalProcessTask
from gcn import get_notice_type, NoticeType
import gcn

from ...import app
from ..core import DispatchHandler

log = get_task_logger(__name__)


def _get_ephemeral_port():
    """Get an ephemeral (unused, high numbered) port. Note that there is
    inherently a race condition in using this function because there is no
    guarantee that the port is still not in use when the caller attempts to use
    it."""
    with socket.socket() as s:
        s.bind(('127.0.0.1', 0))
        _, port = s.getsockname()
    return port


@app.task(base=EternalTask, bind=True, shared=False)
def broker(self):
    """Run an embedded :doc:`Comet VOEvent broker <comet:usage/broker>` to send
    GCNs."""
    # Look up the broker broadcast port from the global configuration object.
    _, _, broadcast_port = app.conf['gcn_broker_address'].partition(':')

    # Pick an ephemeral port to listen to for VOEvent submission.
    # Using an ephemeral port allows us to run multiple instances of
    # GWCelery on the same machine, which may be useful for testing.
    author_port = str(_get_ephemeral_port())
    self.backend.client.set(self.name + '.port', author_port)

    # Create temporary directory for Comet's event database in order to ensure
    # isolation from any other instance of GWCelery on the same machine.
    with tempfile.TemporaryDirectory() as tmpdir:
        # Assemble the command line options for Twisted.
        cmd = [sys.executable, '-m', 'twisted', 'comet', '--verbose',
               '--local-ivo', 'ivo://ligo.org/gwcelery',
               '--receive',
               '--receive-port', author_port,
               '--author-whitelist', '127.0.0.0/8',
               '--eventdb', tmpdir]
        if app.conf['gcn_broker_accept_addresses']:
            cmd += ['--broadcast',
                    '--broadcast-port', broadcast_port,
                    '--broadcast-test-interval', '0']
            for name in app.conf['gcn_broker_accept_addresses']:
                cmd += ['--subscriber-whitelist', socket.gethostbyname(name)]
        subprocess.check_call(cmd)


@app.task(ignore_result=True, shared=False)
def send(message):
    """Send a VOEvent to the local Comet instance for forwarding to GCN.

    Internally, this just calls :doc:`comet-sendvo <comet:usage/publisher>`."""

    # Look up the ephemeral port number saved in Redis.
    port = broker.backend.client.get(broker.name + '.port')

    # Send the VOEvent using comet-sendvo.
    subprocess.run(['comet-sendvo', '--port', port],
                   check=True, input=message, stderr=subprocess.PIPE)


class _VOEventDispatchHandler(DispatchHandler):

    def process_args(self, payload, root):
        notice_type = get_notice_type(root)

        # Just cast to enum for prettier log messages
        try:
            notice_type = NoticeType(notice_type)
        except ValueError:
            pass

        return notice_type, (payload,), {}


handler = _VOEventDispatchHandler()
r"""Function decorator to register a handler callback for specified GCN notice
types. The decorated function is turned into a Celery task, which will be
automatically called whenever a matching GCN notice is received.

Parameters
----------
\*keys
    List of GCN notice types to accept
\*\*kwargs
    Additional keyword arguments for :meth:`celery.Celery.task`.

Examples
--------
Declare a new handler like this::

    @gcn.handler(gcn.NoticeType.FERMI_GBM_GND_POS,
                 gcn.NoticeType.FERMI_GBM_FIN_POS)
    def handle_fermi(payload):
        root = lxml.etree.fromstring(payload)
        # do work here...
"""


def _host_port(address):
    host, port = address.split(':')
    return host, int(port)


@app.task(base=EternalProcessTask if app.conf['gcn_client_address'] else None,
          shared=False)
def listen():
    """Listen to GCN notices forever. GCN notices are dispatched asynchronously
    to tasks that have been registered with
    :meth:`gwcelery.tasks.gcn.handler`."""
    host, port = _host_port(app.conf['gcn_client_address'])
    gcn.listen(host, port, handler=handler.dispatch)



from lxml import etree
from urllib.parse import urlparse

from celery.utils.log import get_task_logger

from . import detchar
from . import gcn
from . import gracedb
from . import ligo_fermi_skymaps
from . import lvalert
from . import raven

log = get_task_logger(__name__)


@gcn.handler(gcn.NoticeType.SNEWS,
             queue='exttrig',
             shared=False)
def handle_sn_gcn(payload):
    """Handles the payload from SNEWS alerts.
    Prepares the alert to be sent to graceDB as 'E' events."""
    root = etree.fromstring(payload)

    #  Get TrigID
    trig_id = root.find("./What/Param[@name='TrigID']").attrib['value']

    event_observatory = 'SNEWS'
    query = 'group: External pipeline: {} grbevent.trigger_id = "{}"'.format(
        event_observatory, trig_id)
    events = gracedb.get_events(query=query)

    if events:
        assert len(events) == 1, 'Found more than one matching GraceDb entry'
        event, = events
        graceid = event['graceid']
        gracedb.replace_event(graceid, payload)

    else:
        graceid = gracedb.create_event(filecontents=payload,
                                       search='Supernova',
                                       group='External',
                                       pipeline=event_observatory)
        event = gracedb.get_event(graceid)
        start = event['gpstime'] - 10
        end = start + 10
        detchar.check_vectors(event, event['graceid'], start, end)


@gcn.handler(gcn.NoticeType.FERMI_GBM_ALERT,
             gcn.NoticeType.FERMI_GBM_FLT_POS,
             gcn.NoticeType.FERMI_GBM_GND_POS,
             gcn.NoticeType.FERMI_GBM_FIN_POS,
             gcn.NoticeType.SWIFT_BAT_GRB_ALERT,
             gcn.NoticeType.SWIFT_BAT_GRB_LC,
             queue='exttrig',
             shared=False)
def handle_grb_gcn(payload):
    """Handles the payload from Fermi and Swift alerts.
    Prepares the alert to be sent to graceDB as 'E' events."""
    root = etree.fromstring(payload)
    u = urlparse(root.attrib['ivorn'])
    stream_path = u.path

    #  Get TrigID
    trig_id = root.find("./What/Param[@name='TrigID']").attrib['value']

    stream_obsv_dict = {'/SWIFT': 'Swift',
                        '/Fermi': 'Fermi'}
    event_observatory = stream_obsv_dict[stream_path]
    query = 'group: External pipeline: {} grbevent.trigger_id = "{}"'.format(
        event_observatory, trig_id)
    events = gracedb.get_events(query=query)

    if events:
        assert len(events) == 1, 'Found more than one matching GraceDb entry'
        event, = events
        graceid = event['graceid']
        gracedb.replace_event(graceid, payload)

    else:
        graceid = gracedb.create_event(filecontents=payload,
                                       search='GRB',
                                       group='External',
                                       pipeline=event_observatory)
        event = gracedb.get_event(graceid)
        start = event['gpstime']
        end = start + event['extra_attributes']['GRB']['trigger_duration']
        detchar.check_vectors(event, event['graceid'], start, end)


@lvalert.handler('superevent',
                 'mdc_superevent',
                 'test_superevent',
                 'external_fermi',
                 'external_fermi_grb',
                 'external_grb',
                 'external_swift',
                 shared=False)
def handle_grb_lvalert(alert):
    """Parse an LVAlert message related to superevents/GRB external triggers
    and dispatch it to other tasks.

    Notes
    -----

    This LVAlert message handler is triggered by creating a new superevent or
    GRB external trigger event, or applying the ``EM_COINC`` label to any
    superevent:

    * Any new event triggers a coincidence search with
      :meth:`gwcelery.tasks.raven.coincidence_search`.
    * The ``EM_COINC`` label triggers the creation of a combined GW-GRB sky map
      using :meth:`gwcelery.tasks.ligo_fermi_skymaps.create_combined_skymap`.
    """
    # Determine GraceDb ID
    graceid = alert['uid']

    if alert['alert_type'] == 'new' and \
            alert['object'].get('group', '') == 'External':
        raven.coincidence_search(graceid, alert['object'], group='CBC').delay()
        raven.coincidence_search(graceid, alert['object'],
                                 group='Burst').delay()
    elif graceid.startswith('S'):
        preferred_event_id = gracedb.get_superevent(graceid)['preferred_event']
        group = gracedb.get_event(preferred_event_id)['group']
        if alert['alert_type'] == 'new':
            raven.coincidence_search(graceid, alert['object'],
                                     group=group,
                                     pipelines=['Fermi', 'Swift']).delay()
        elif alert['alert_type'] == 'label_added':
            if alert['data']['name'] == 'EM_COINC':
                ligo_fermi_skymaps.create_combined_skymap(graceid).delay()
                raven.calculate_coincidence_far(graceid, group).delay()
                raven.calculate_spacetime_coincidence_far(graceid,
                                                          group).delay()


@lvalert.handler('superevent',
                 'mdc_superevent',
                 'test_superevent',
                 'external_snews',
                 'external_snews_supernova',
                 shared=False)
def handle_sn_lvalert(alert):
    """Parse an LVAlert message related to superevents/SN external triggers and
    dispatch it to other tasks.

    Notes
    -----

    This LVAlert message handler is triggered by creating a new superevent or
    SN external trigger event, or applying the ``EM_COINC`` label to any
    superevent:

    * Any new event triggers a coincidence search with
      :meth:`gwcelery.tasks.raven.coincidence_search`.
    """
    # Determine GraceDb ID
    graceid = alert['uid']

    if alert['alert_type'] == 'new' and \
            alert['object'].get('group', '') == 'External':
        raven.coincidence_search(graceid, alert['object'],
                                 group='Burst', pipelines=['SNEWS']).delay()
    elif graceid.startswith('S'):
        preferred_event_id = gracedb.get_superevent(graceid)['preferred_event']
        group = gracedb.get_event(preferred_event_id)['group']
        if alert['alert_type'] == 'new' and group == 'Burst':
            raven.coincidence_search(graceid, alert['object'],
                                     group=group, pipelines=['SNEWS']).delay()



"""Application configuration for ``gracedb.ligo.org``. Inherits all settings
from :mod:`gwcelery.conf.playground`, with the exceptions below."""

from . import *  # noqa: F401, F403

lvalert_host = 'lvalert.cgca.uwm.edu'
"""LVAlert host."""

gracedb_host = 'gracedb.ligo.org'
"""GraceDb host."""

gcn_broker_accept_addresses = ['capella2.gsfc.nasa.gov']
"""List of hosts from which the broker will accept connections.
If empty, then completely disable the broker's broadcast capability."""

gcn_client_address = '68.169.57.253:8096'
"""The VOEvent listener will connect to this address to receive GCNs.
If empty, then completely disable the GCN listener.

We are temporarily using the pre-registered port 8096 for receiving
proprietary LIGO/Virgo alerts on emfollow.ligo.caltech.edu. This means that
the capability to receive GCNs requires setting up a site configuration in
advance with Scott Barthelmey.

Once we switch to sending public alerts exclusively, then we can switch
back to using port 8099 for anonymous access, requiring no prior site
configuration."""

low_latency_frame_types = {'H1': 'H1_llhoft',
                           'L1': 'L1_llhoft',
                           'V1': 'V1_llhoft'}
"""Types of frames used in Parameter Estimation with LALInference (see
:mod:`gwcelery.tasks.lalinference`)"""

high_latency_frame_types = {'H1': 'H1_HOFT_C00',
                            'L1': 'L1_HOFT_C00',
                            'V1': 'V1Online'}
"""Types of nonllhoft-frames used in Parameter Estimation with LALInference
(see :mod:`gwcelery.tasks.lalinference`)"""

strain_channel_names = {'H1': 'H1:GDS-CALIB_STRAIN',
                        'L1': 'L1:GDS-CALIB_STRAIN',
                        'V1': 'V1:Hrec_hoft_16384Hz'}
"""Names of h(t) channels used in Parameter Estimation with LALInference (see
:mod:`gwcelery.tasks.lalinference`)"""

sentry_environment = 'production'
"""Record this `environment tag
<https://docs.sentry.io/enriching-error-data/environments/>`)` in Sentry log
messages."""



"""Base classes for other Celery tasks."""
from operator import itemgetter

from celery import group
from celery.utils.log import get_task_logger

from .. import app

log = get_task_logger(__name__)


@app.task(shared=False)
def identity(arg):
    """Identity task (returns its input)."""
    return arg


@app.task(shared=False)
def _pack(result, i):
    return i, result


@app.task(shared=False)
def _unpack(results):
    return tuple(map(itemgetter(1), sorted(results, key=itemgetter(0))))


def ordered_group(*args):
    """Like :meth:`celery.group`, but preserve order.

    This is a temporary workaround for
    https://github.com/celery/celery/pull/4858."""
    return group(arg | _pack.s(i) for i, arg in enumerate(args)) | _unpack.s()


class DispatchHandler(dict):

    def process_args(self, *args, **kwargs):
        r"""Determine key and callback arguments.

        The default implementation treats the first positional argument as the
        key.

        Parameters
        ----------
        \*args
            Arguments passed to :meth:`__call__`.
        \*\*kwargs
            Keyword arguments passed to :meth:`__call__`.

        Returns
        -------
        key
            The key to determine which callback to invoke.
        \*args
            The arguments to pass to the registered callback.
        \*\*kwargs
            The keyword arguments to pass to the registered callback.
        """
        key, *args = args
        return key, args, kwargs

    def __call__(self, *keys, **kwargs):
        r"""Create a new task and register it as a callback for handling the
        given keys.

        Parameters
        ----------
        \*keys : list
            Keys to match
        \*\*kwargs
            Additional keyword arguments for `celery.Celery.task`.
        """

        def wrap(f):
            f = app.task(ignore_result=True, **kwargs)(f)
            for key in keys:
                self.setdefault(key, []).append(f)
            return f

        return wrap

    def dispatch(self, *args, **kwargs):
        log.debug('considering dispatch: args=%r, kwargs=%r', args, kwargs)
        try:
            key, args, kwargs = self.process_args(*args, **kwargs)
        except (TypeError, ValueError):
            log.exception('error unpacking key')
            return
        log.debug('unpacked: key=%r, args=%r, kwargs=%r', key, args, kwargs)

        try:
            matching_handlers = self[key]
        except KeyError:
            log.warning('ignoring unrecognized key: %r', key)
        else:
            log.info('calling handlers %r for key %r', matching_handlers, key)
            group([handler.s() for handler in matching_handlers]).apply_async(
                args, kwargs)



"""Module containing the computation of p_astro by source category
   See https://dcc.ligo.org/LIGO-T1800072 for details.
"""
import io
import json

from celery.utils.log import get_task_logger
from glue.ligolw import ligolw
from glue.ligolw.ligolw import LIGOLWContentHandler
from glue.ligolw import array as ligolw_array
from glue.ligolw import param as ligolw_param
from glue.ligolw import utils as ligolw_utils
from glue.ligolw import lsctables
from lal import rate
import numpy as np

from ..import app

from . import p_astro_other

log = get_task_logger(__name__)

# adapted from gstlal far.py RankingStatPDF


class _RankingStatPDF(object):
    ligo_lw_name_suffix = "gstlal_inspiral_rankingstatpdf"

    @classmethod
    def from_xml(cls, xml, name):
        """
        Find the root of the XML tree containing the
        serialization of this object
        """
        xml, = [elem for elem in
                xml.getElementsByTagName(ligolw.LIGO_LW.tagName)
                if elem.hasAttribute("Name") and
                elem.Name == "%s:%s" % (name, cls.ligo_lw_name_suffix)]
        # create a uninitialized instance
        self = super().__new__(cls)
        # populate from XML
        self.noise_lr_lnpdf = rate.BinnedLnPDF.from_xml(xml, "noise_lr_lnpdf")
        self.signal_lr_lnpdf = rate.BinnedLnPDF.from_xml(xml,
                                                         "signal_lr_lnpdf")
        self.zero_lag_lr_lnpdf = rate.BinnedLnPDF.from_xml(
            xml, "zero_lag_lr_lnpdf")
        return self


def _parse_likelihood_control_doc(xmldoc):
    name = "gstlal_inspiral_likelihood"
    rankingstatpdf = _RankingStatPDF.from_xml(xmldoc, name)
    if rankingstatpdf is None:
        raise ValueError("document does not contain likelihood ratio data")
    return rankingstatpdf


@ligolw_array.use_in
@ligolw_param.use_in
@lsctables.use_in
class _ContentHandler(LIGOLWContentHandler):
    pass


def _get_ln_f_over_b(ranking_data_bytes, ln_likelihood_ratios):
    ranking_data_xmldoc, _ = ligolw_utils.load_fileobj(
        io.BytesIO(ranking_data_bytes), contenthandler=_ContentHandler)
    rankingstatpdf = _parse_likelihood_control_doc(ranking_data_xmldoc)
    # affect the zeroing of the PDFs below threshold by hacking the
    # histograms. Do the indexing ourselves to not 0 the bin @ threshold
    ln_likelihood_ratio_threshold = \
        app.conf['p_astro_gstlal_ln_likelihood_threshold']
    noise_lr_lnpdf = rankingstatpdf.noise_lr_lnpdf
    rankingstatpdf.noise_lr_lnpdf.array[
        :noise_lr_lnpdf.bins[0][ln_likelihood_ratio_threshold]] \
        = 0.
    rankingstatpdf.noise_lr_lnpdf.normalize()
    signal_lr_lnpdf = rankingstatpdf.signal_lr_lnpdf
    rankingstatpdf.signal_lr_lnpdf.array[
        :signal_lr_lnpdf.bins[0][ln_likelihood_ratio_threshold]] \
        = 0.
    rankingstatpdf.signal_lr_lnpdf.normalize()
    zero_lag_lr_lnpdf = rankingstatpdf.zero_lag_lr_lnpdf
    rankingstatpdf.zero_lag_lr_lnpdf.array[
        :zero_lag_lr_lnpdf.bins[0][ln_likelihood_ratio_threshold]] \
        = 0.
    rankingstatpdf.zero_lag_lr_lnpdf.normalize()

    f = rankingstatpdf.signal_lr_lnpdf
    b = rankingstatpdf.noise_lr_lnpdf
    ln_f_over_b = \
        np.array([f[ln_lr, ] - b[ln_lr, ] for ln_lr in ln_likelihood_ratios])
    if np.isnan(ln_f_over_b).any():
        raise ValueError("NaN encountered in ranking statistic PDF ratios")
    if np.isinf(np.exp(ln_f_over_b)).any():
        raise ValueError(
            "infinity encountered in ranking statistic PDF ratios")
    return ln_f_over_b


def _get_event_ln_likelihood_ratio_svd_endtime_mass(coinc_bytes):
    coinc_xmldoc, _ = ligolw_utils.load_fileobj(
        io.BytesIO(coinc_bytes), contenthandler=_ContentHandler)
    coinc_event, = lsctables.CoincTable.get_table(coinc_xmldoc)
    coinc_inspiral, = lsctables.CoincInspiralTable.get_table(coinc_xmldoc)
    sngl_inspiral = lsctables.SnglInspiralTable.get_table(coinc_xmldoc)

    assert all([sngl_inspiral[i].Gamma0 == sngl_inspiral[i+1].Gamma0
                for i in range(len(sngl_inspiral)-1)]), \
        "svd bank different between ifos!"
    return (coinc_event.likelihood,
            coinc_inspiral.end_time,
            coinc_inspiral.mass,
            sngl_inspiral[0].mass1,
            sngl_inspiral[0].mass2,
            coinc_inspiral.snr,
            coinc_inspiral.combined_far)


# This is the function that computes p-astro for a new
# event using mean values of Poisson expected counts
# constructed from all the previous events. This is
# the function that will be invoked with every new
# GraceDB entry


def p_astro_update(category, event_bayesfac_dict, mean_values_dict):

    if category == "Terr":
        numerator = mean_values_dict["Terr"]
    else:
        numerator = \
            event_bayesfac_dict[category]*mean_values_dict[category]

    denominator = mean_values_dict["Terr"] + \
        np.sum([mean_values_dict[key]*event_bayesfac_dict[key]
                for key in event_bayesfac_dict.keys()])

    return numerator/denominator


@app.task(shared=False)
def compute_p_astro(files):
    """
    Task to compute `p_astro` by source category.

    Parameters
    ----------
    files : tuple
        Tuple of byte content from (coinc.xml, ranking_data.xml.gz)

    Returns
    -------
    p_astros : str
        JSON dump of the p_astro by source category

    Example
    -------
    >>> p_astros = json.loads(compute_p_astro(files))
    >>> p_astros
    {'BNS': 0.999, 'BBH': 0.0, 'NSBH': 0.0, 'Terr': 0.001}
    """
    coinc_bytes, ranking_data_bytes = files

    # Acquire information pertaining to the event from coinc.xml
    # uploaded to GraceDB
    log.info(
        'Fetching ln_likelihood_ratio, svd bin, endtime, mass from coinc.xml')
    event_ln_likelihood_ratio, event_endtime, \
        event_mass, event_mass1, event_mass2, snr, far = \
        _get_event_ln_likelihood_ratio_svd_endtime_mass(coinc_bytes)

    # Using the zerolag log likelihood ratio value event,
    # and the foreground/background model information provided
    # in ranking_data.xml.gz, compute the ln(f/b) value for this event
    zerolag_ln_likelihood_ratios = np.array([event_ln_likelihood_ratio])
    log.info('Computing f_over_b from ranking_data.xml.gz')
    # resort to approximate computation if NaN encountered
    try:
        ln_f_over_b = _get_ln_f_over_b(ranking_data_bytes,
                                       zerolag_ln_likelihood_ratios)
    except ValueError:
        log.warning("NaN encountered, using approximate method...")
        return p_astro_other.compute_p_astro(snr,
                                             far,
                                             event_mass1,
                                             event_mass2)
    else:
        num_bins = 3
        mean_bns = 2.07613829518
        mean_nsbh = 1.65747751787
        mean_bbh = 11.8941873831
        mean_terr = 3923

        a_hat_bns = int(event_mass1 <= 3 and event_mass2 <= 3)
        a_hat_bbh = int(event_mass2 > 3 and event_mass2 > 3)
        a_hat_nsbh = int(min([event_mass1, event_mass2]) <= 3 and
                         max([event_mass1, event_mass2]) > 3)

        # Fix mean values (1 per source category) from
        mean_values_dict = {"BNS": mean_bns,
                            "NSBH": mean_nsbh,
                            "BBH": mean_bbh,
                            "Terr": mean_terr}

        # These are the bayes factor values that need to be
        # constructed with every new event/GraceDB upload
        rescaled_fb = num_bins*np.exp(ln_f_over_b)[0]
        bns_bayesfac = a_hat_bns*rescaled_fb
        nsbh_bayesfac = a_hat_nsbh*rescaled_fb
        bbh_bayesfac = a_hat_bbh*rescaled_fb
        event_bayesfac_dict = {"BNS": bns_bayesfac,
                               "NSBH": nsbh_bayesfac,
                               "BBH": bbh_bayesfac}

        # Compute the p-astro values for each source category
        # using the mean values
        p_astro_values = {}
        for category in mean_values_dict:
            p_astro_values[category] = \
                p_astro_update(category=category,
                               event_bayesfac_dict=event_bayesfac_dict,
                               mean_values_dict=mean_values_dict)
        return json.dumps(p_astro_values)



"""Generate and upload automated circulars."""
import ligo.followup_advocate

from . import gracedb


@gracedb.task(shared=False)
def create_circular(graceid):
    """Create and return circular txt."""
    return ligo.followup_advocate.compose(graceid, client=gracedb.client)



"""A `Nagios plugin <https://nagios-plugins.org/doc/guidelines.html>`_
for monitoring GWCelery."""
from enum import IntEnum
from sys import exit
from traceback import format_exc, format_exception

from celery.bin.base import Command
from celery_eternal import EternalTask
import kombu.exceptions
import sleek_lvalert

# Make sure that all tasks are registered
from . import tasks


class NagiosPluginStatus(IntEnum):
    """Nagios plugin status codes."""
    OK = 0
    WARNING = 1
    CRITICAL = 2
    UNKNOWN = 3


class NagiosCriticalError(Exception):
    """An exception that maps to a Nagios status of `CRITICAL`."""


def get_active_queues(inspector):
    return {queue['name']
            for queues in (inspector.active_queues() or {}).values()
            for queue in queues}


def get_active_tasks(inspector):
    return {task['name']
            for tasks in inspector.active().values()
            for task in tasks}


def get_active_lvalert_nodes(app):
    client = sleek_lvalert.LVAlertClient(server=app.conf['lvalert_host'])
    client.connect()
    client.process(block=False)
    active = set(client.get_subscriptions())
    client.disconnect()
    return active


def get_expected_queues(app):
    # Get the queues for all registered tasks.
    result = {getattr(task, 'queue', None) for task in app.tasks.values()}
    # We use 'celery' for all tasks that do not explicitly specify a queue.
    result -= {None}
    result |= {'celery'}
    # Done.
    return result


def get_expected_tasks(app):
    return {name for name, task in app.tasks.items()
            if isinstance(task, EternalTask)}


def get_expected_lvalert_nodes():
    return set(tasks.lvalert.handler.keys())


def check_status(app):
    connection = app.connection()
    try:
        connection.ensure_connection(max_retries=1)
    except kombu.exceptions.OperationalError as e:
        raise NagiosCriticalError('No connection to broker') from e

    inspector = app.control.inspect()

    active = get_active_queues(inspector)
    expected = get_expected_queues(app)
    missing = expected - active
    if missing:
        raise NagiosCriticalError('Not all expected queues are active') from \
              AssertionError('Missing queues: ' + ', '.join(missing))

    active = get_active_tasks(inspector)
    expected = get_expected_tasks(app)
    missing = expected - active
    if missing:
        raise NagiosCriticalError('Not all expected tasks are active') from \
              AssertionError('Missing tasks: ' + ', '.join(missing))

    active = get_active_lvalert_nodes(app)
    expected = get_expected_lvalert_nodes()
    missing = expected - active
    extra = active - expected
    if missing:
        raise NagiosCriticalError('Not all lvalert nodes are subscribed') \
            from AssertionError('Missing nodes: ' + ', '.join(missing))
    if extra:
        raise NagiosCriticalError('Too many lvalert nodes are subscribed') \
            from AssertionError('Extra nodes: ' + ', '.join(extra))


class NagiosCommand(Command):
    """Check Celery status for monitoring with Nagios."""

    def run(self, **kwargs):
        try:
            check_status(self.app)
        except NagiosCriticalError as e:
            status = NagiosPluginStatus.CRITICAL
            output, = e.args
            e = e.__cause__
            detail = ''.join(format_exception(type(e), e, e.__traceback__))
        except:  # noqa: E722
            status = NagiosPluginStatus.UNKNOWN
            output = 'Unexpected error'
            detail = format_exc()
        else:
            status = NagiosPluginStatus.OK
            output = 'Running normally'
            detail = None
        print('{}: {}'.format(status.name, output))
        if detail:
            print(detail)
        exit(status)



"""Tasks that comprise the alert orchestrator, which responsible for the
vetting and annotation workflow to produce preliminary, initial, and update
alerts for gravitational-wave event candidates."""
import json
import re
from urllib.error import URLError

from celery import chain, group
from ligo.gracedb.rest import HTTPError

from ..import app
from . import bayestar
from . import circulars
from .core import identity, ordered_group
from . import detchar
from . import em_bright
from . import gcn
from . import gracedb
from . import lalinference
from . import lvalert
from . import skymaps
from . import p_astro_gstlal, p_astro_other


@lvalert.handler('superevent',
                 'mdc_superevent',
                 shared=False)
def handle_superevent(alert):
    """Schedule annotations for new superevents.

    After waiting for a time specified by the
    :obj:`~gwcelery.conf.orchestrator_timeout` configuration variable
    for the choice of preferred event to settle down, this task peforms data
    quality checks with :meth:`gwcelery.tasks.detchar.check_vectors` and
    calls :meth:`~gwcelery.tasks.orchestrator.preliminary_alert` to send a
    preliminary GCN notice.
    """

    superevent_id = alert['uid']

    if alert['alert_type'] == 'new':
        start = alert['object']['t_start']
        end = alert['object']['t_end']

        gracedb.create_label.s('ADVREQ', superevent_id).apply_async()

        (
            _get_preferred_event.si(superevent_id).set(
                countdown=app.conf['orchestrator_timeout']
            )
            |
            gracedb.get_event.s()
            |
            group(
                detchar.check_vectors.s(superevent_id, start, end)
                |
                preliminary_alert.s(superevent_id),

                parameter_estimation.s(superevent_id)
            )
        ).apply_async()
    # check DQV label on superevent, run check_vectors if required
    elif alert['alert_type'] == 'event_added':
        # FIXME Getting the graceid from the most recent event added log
        # this is a temporary solution, fix post ER13
        regex = re.compile(r'Added event: ([GMT][0-9]+)')
        for log in reversed(gracedb.get_log(superevent_id)):
            match = regex.match(log['comment'])
            if match:
                new_event_id = match[1]
                break
        else:
            raise ValueError('Added event log message not found')

        start = alert['data']['t_start']
        end = alert['data']['t_end']

        if 'DQV' in gracedb.get_labels(superevent_id):
            (
                detchar.check_vectors.s(new_event_id, superevent_id,
                                        start, end)
                |
                _update_if_dqok.si(superevent_id, new_event_id)
            ).apply_async()

    elif alert['alert_type'] == 'label_added':
        label_name = alert['data']['name']
        if label_name == 'ADVOK':
            initial_alert(superevent_id)
        elif label_name == 'ADVNO':
            retraction_alert(superevent_id)


@lvalert.handler('cbc_gstlal',
                 'cbc_spiir',
                 'cbc_pycbc',
                 'cbc_mbtaonline',
                 shared=False)
def handle_cbc_event(alert):
    """Peform annotations for CBC events that depend on pipeline-specific
    matched-filter parameter estimates.

    Notes
    -----

    This LVAlert message handler is triggered by updates that include the files
    ``psd.xml.gz`` and ``ranking_data.xml.gz``. The table below lists which
    files are created as a result, and which tasks generate them.

    ============================== =====================================================
    File                           Task
    ============================== =====================================================
    ``bayestar.fits``              :meth:`gwcelery.tasks.bayestar.localize`
    ``source_classification.json`` :meth:`gwcelery.tasks.em_bright.classifier`
    ``p_astro.json``               :meth:`gwcelery.tasks.p_astro_gstlal.compute_p_astro`
    ============================== =====================================================
    """  # noqa: E501

    graceid = alert['uid']

    # p_astro calculation for other pipelines
    if alert['alert_type'] == 'new' \
            and (alert['object']['pipeline'].lower() != 'gstlal'
                 or alert['object']['search'] == 'MDC'):
        extra_attributes = alert['object']['extra_attributes']
        snr = extra_attributes['CoincInspiral']['snr']
        far = alert['object']['far']
        mass1 = extra_attributes['SingleInspiral'][0]['mass1']
        mass2 = extra_attributes['SingleInspiral'][0]['mass2']
        (
            p_astro_other.compute_p_astro.s(snr, far, mass1, mass2)
            |
            gracedb.upload.s(
                'p_astro.json', graceid,
                'p_astro computation complete', ['p_astro', 'public']
            )
            |
            gracedb.create_label.si('PASTRO_READY', graceid)
        ).delay()

    if alert['alert_type'] != 'log':
        return

    filename = alert['data']['filename']

    if filename == 'psd.xml.gz':
        (
            ordered_group(
                gracedb.download.s('coinc.xml', graceid),
                gracedb.download.s('psd.xml.gz', graceid)
            )
            |
            group(
                bayestar.localize.s(graceid)
                |
                gracedb.upload.s(
                    'bayestar.fits', graceid,
                    'sky localization complete', ['sky_loc', 'public']
                )
                |
                gracedb.create_label.si('SKYMAP_READY', graceid),

                em_bright.classifier.s(graceid)
                |
                gracedb.upload.s(
                    'source_classification.json', graceid,
                    'source classification complete', ['em_bright', 'public']
                )
                |
                gracedb.create_label.si('EMBRIGHT_READY', graceid)
            )
        ).delay()
    elif filename == 'ranking_data.xml.gz':
        (
            gracedb.get_event.s(graceid)
            |
            ordered_group(
                gracedb.download.si('coinc.xml', graceid),
                gracedb.download.si('ranking_data.xml.gz', graceid)
            )
            |
            p_astro_gstlal.compute_p_astro.s()
            |
            gracedb.upload.s(
                'p_astro.json', graceid,
                'p_astro computation complete', ['p_astro', 'public']
            )
            |
            gracedb.create_label.si('PASTRO_READY', graceid)
        ).delay()


@app.task(autoretry_for=(HTTPError, URLError, TimeoutError),
          default_retry_delay=20.0, retry_backoff=True,
          retry_kwargs=dict(max_retries=500), shared=False)
def _download(*args, **kwargs):
    """Download a file from GraceDb.

    This works just like :func:`gwcelery.tasks.gracedb.download`, except that
    it is retried for both :class:`TimeoutError` and
    :class:`~urllib.error.URLError`. In particular, it will be retried for 404
    (not found) errors."""
    return gracedb.download(*args, **kwargs)


@app.task(shared=False, ignore_result=True)
def _update_if_dqok(superevent_id, event_id):
    """Update `preferred_event` of `superevent_id` to `event_id`
    if `DQOK` label has been applied
    """
    if 'DQOK' in gracedb.get_labels(superevent_id):
        gracedb.update_superevent(superevent_id, preferred_event=event_id)
        gracedb.create_log(
            "DQOK applied based on new event %s" % (event_id), superevent_id)


@gracedb.task(shared=False)
def _get_preferred_event(superevent_id):
    """Determine preferred event for a superevent by querying GraceDb.

    This works just like :func:`gwcelery.tasks.gracedb.get_superevent`, except
    that it returns only the preferred event, and not the entire GraceDb JSON
    response."""
    return gracedb.get_superevent(superevent_id)['preferred_event']


@gracedb.task(shared=False)
def _create_voevent(classification, *args, **kwargs):
    r"""Create a VOEvent record from an EM bright JSON file.

    Parameters
    ----------
    classification : tuple, None
        A collection of JSON strings, generated by
        :meth:`gwcelery.tasks.em_bright.classifier` and
        :meth:`gwcelery.tasks.p_astro_gstlal.compute_p_astro` respectively; or
        None
    \*args
        Additional positional arguments passed to
        :meth:`gwcelery.tasks.gracedb.create_voevent`.
    \*\*kwargs
        Additional keyword arguments passed to
        :meth:`gwcelery.tasks.gracedb.create_voevent`.

    Returns
    -------
    str
        The filename of the newly created VOEvent.
    """
    kwargs = dict(kwargs)

    if classification is not None:
        data = {}
        for text in classification:
            data.update(json.loads(text))

        try:
            kwargs['ProbHasNS'] = 0.01 * data['Prob NS2']
        except KeyError:
            pass

        try:
            kwargs['ProbHasRemnant'] = 0.01 * data['Prob EMbright']
        except KeyError:
            pass

        try:
            kwargs['BNS'] = data['BNS']
        except KeyError:
            pass

        try:
            kwargs['NSBH'] = data['NSBH']
        except KeyError:
            pass

        try:
            kwargs['BBH'] = data['BBH']
        except KeyError:
            pass

        try:
            kwargs['Terrestrial'] = data['Terr']
        except KeyError:
            pass

    skymap_filename = kwargs.get('skymap_filename')
    if skymap_filename is not None:
        skymap_type = re.sub(r'\.fits(\..+)?$', '', skymap_filename)
        kwargs.setdefault('skymap_type', skymap_type)

    return gracedb.create_voevent(*args, **kwargs)


@app.task(ignore_result=True, shared=False)
def preliminary_alert(event, superevent_id):
    """Produce a preliminary alert by copying any sky maps.

    This consists of the following steps:

    1.   Copy any sky maps and source classification from the preferred event
         to the superevent.
    2.   Create standard annotations for sky maps including all-sky plots by
         calling :meth:`gwcelery.tasks.skymaps.annotate_fits`.
    3.   Create a preliminary VOEvent.
    4.   Send the VOEvent to GCN.
    5.   Apply the GCN_PRELIM_SENT label to the superevent.
    6.   Create and upload a GCN Circular draft.
    """
    preferred_event_id = event['graceid']

    if event['group'] == 'CBC':
        skymap_filename = 'bayestar.fits'
    elif event['pipeline'] == 'CWB':
        skymap_filename = 'cWB.fits.gz'
    elif event['pipeline'] == 'oLIB':
        skymap_filename = 'oLIB.fits.gz'
    else:
        skymap_filename = None

    original_skymap_filename = skymap_filename
    if skymap_filename.endswith('.fits'):
        skymap_filename += '.gz'

    # Make the event public.
    # FIXME: For ER13, only expose mock events.
    if event['search'] == 'MDC':
        canvas = gracedb.expose.s(superevent_id)
    else:
        canvas = chain()

    # If there is a sky map, then copy it to the superevent and create plots.
    if skymap_filename is not None:
        canvas |= (
            _download.si(original_skymap_filename, preferred_event_id)
            |
            group(
                gracedb.upload.s(
                    original_skymap_filename,
                    superevent_id,
                    message='Localization copied from {}'.format(
                        preferred_event_id),
                    tags=['sky_loc', 'public']
                )
                |
                _download.si(original_skymap_filename, superevent_id)
                |
                skymaps.flatten.s(skymap_filename)
                |
                gracedb.upload.s(
                    skymap_filename,
                    superevent_id,
                    message='Flattened from multiresolution file {}'.format(
                        original_skymap_filename),
                    tags=['sky_loc', 'public']
                )
                |
                gracedb.create_label.si('SKYMAP_READY', superevent_id),
            )
        )

        (
            _download.si(original_skymap_filename, superevent_id)
            |
            skymaps.annotate_fits(
                original_skymap_filename,
                superevent_id,
                ['sky_loc', 'public']
            )
        ).apply_async()

    # If this is a CBC event, then copy the EM bright classification.
    if event['group'] == 'CBC':
        canvas |= group(
            _download.si('source_classification.json', preferred_event_id)
            |
            gracedb.upload.s(
                'source_classification.json',
                superevent_id,
                message='Source properties copied from {}'.format(
                    preferred_event_id),
                tags=['em_bright', 'public']
            )
            |
            gracedb.create_label.si('EMBRIGHT_READY', superevent_id)
            |
            _download.si('source_classification.json', superevent_id),

            _download.si('p_astro.json', preferred_event_id)
            |
            gracedb.upload.s(
                'p_astro.json',
                superevent_id,
                message='Source classification copied from {}'.format(
                    preferred_event_id),
                tags=['p_astro', 'public']
            )
            |
            gracedb.create_label.si('PASTRO_READY', superevent_id)
            |
            _download.si('p_astro.json', superevent_id)
        ) | identity.s()  # FIXME: necessary to pass result to next task?
    else:
        canvas |= identity.si(None)

    # Send GCN notice and upload GCN circular draft for online events.
    trials_factor = \
        app.conf['preliminary_alert_trials_factor'][event['group'].lower()]
    if not event['offline'] \
            and trials_factor * event['far'] <= \
            app.conf[
                'preliminary_alert_far_threshold'][event['group'].lower()] \
            and 'DQV' not in gracedb.get_labels(superevent_id):
        canvas |= (
            _create_voevent.s(
                superevent_id, 'preliminary',
                skymap_filename=skymap_filename,
                # FIXME: for ER13, only send public alerts for MDC events.
                internal=(event['search'] != 'MDC'),
                open_alert=True
            )
            |
            group(
                gracedb.download.s(superevent_id)
                |
                gcn.send.s()
                |
                gracedb.create_label.si('GCN_PRELIM_SENT', superevent_id),

                # FIXME: after ER13, we need to add the 'public' tag to make
                # the VOEvent file public in GraceDb, like we do for the
                # initial, update, and retraction alerts.

                circulars.create_circular.si(superevent_id)
                |
                gracedb.upload.s(
                    'circular.txt',
                    superevent_id,
                    'Automatically generated draft of initial GCN Circular',
                    tags=['em_follow']
                )
            )
        )

    canvas.apply_async()


@app.task(ignore_result=True, shared=False)
def parameter_estimation(event, superevent_id):
    """Tasks for Parameter Estimation Followup with LALInference

    This consists of the following steps:

    1.   Upload an ini file which is suitable for the target event.
    2.   Start Parameter Estimation if FAR is smaller than the PE threshold.
    """
    preferred_event_id = event['graceid']
    # FIXME: it will be better to start parater estimation for 'burst' events.
    if event['group'] == 'CBC' and event['search'] != 'MDC':
        canvas = lalinference.prepare_ini.s(event, superevent_id)
        next_task = gracedb.upload.s(
                        filename=lalinference.ini_name,
                        graceid=superevent_id,
                        message='Automatically generated LALInference ' +
                                'configuration file for this event.',
                        tags='pe'
                    )
        if event['far'] <= app.conf['pe_threshold']:
            next_task = group(
                next_task,

                lalinference.start_pe.s(preferred_event_id, superevent_id)
            )
        canvas |= next_task

        canvas.apply_async()


@app.task(ignore_result=True, shared=False)
def initial_or_update_alert(superevent_id, alert_type, skymap_filename=None,
                            source_classification_filename=None,
                            p_astro_filename=None):
    """
    Create and send initial or update GCN notice.

    Parameters
    ----------
    superevent_id : str
        The superevent ID.
    alert_type : {'initial', 'update'}
        The alert type.
    skymap_filename : str, optional
        The sky map to send.
        If None, then most recent public sky map is used.
    source_classification_filename : str, optional
        The source classification file to use.
        If None, then most recent one is used.
    p_astro_filename : str, optional
        The p_astro file to use.
        If None, then most recent one is used.
    """
    skymap_needed = (skymap_filename is None)
    source_classification_needed = (source_classification_filename is None)
    p_astro_needed = (p_astro_filename is None)
    if skymap_needed or source_classification_needed or p_astro_needed:
        for message in gracedb.get_log(superevent_id):
            t = message['tag_names']
            f = message['filename']
            if not f:
                continue
            if skymap_needed \
                    and {'sky_loc', 'public'}.issubset(t) \
                    and (f.endswith('.fits') or f.endswith('.fits.gz')):
                skymap_filename = f
            if source_classification_needed \
                    and 'em_bright' in t \
                    and f.endswith('.json'):
                source_classification_filename = f
            if p_astro_needed \
                    and 'p_astro' in t \
                    and f.endswith('.json'):
                p_astro_filename = f

    (
        gracedb.expose.s(superevent_id)
        |
        ordered_group(
            gracedb.download.si(source_classification_filename, superevent_id),
            gracedb.download.si(p_astro_filename, superevent_id)
        )
        |
        _create_voevent.s(
            superevent_id,
            alert_type,
            skymap_filename=skymap_filename,
            internal=False,
            open_alert=True,
            vetted=True
        )
        |
        group(
            gracedb.download.s(superevent_id)
            |
            gcn.send.s(),

            gracedb.create_tag.s('public', superevent_id)
        )
    ).apply_async()


@app.task(ignore_result=True, shared=False)
def initial_alert(superevent_id, skymap_filename=None,
                  source_classification_filename=None, p_astro_filename=None):
    """Produce an initial alert.

    This does nothing more than call
    :meth:`~gwcelery.tasks.orchestrator.initial_or_update_alert` with
    ``alert_type='initial'``.

    Parameters
    ----------
    superevent_id : str
        The superevent ID.
    skymap_filename : str, optional
        The sky map to send.
        If None, then most recent public sky map is used.
    source_classification_filename : str, optional
        The source classification file to use.
        If None, then most recent one is used.
    p_astro_filename : str, optional
        The p_astro file to use.
        If None, then most recent one is used.
    """
    initial_or_update_alert(superevent_id, 'initial', skymap_filename,
                            source_classification_filename, p_astro_filename)


@app.task(ignore_result=True, shared=False)
def update_alert(superevent_id, skymap_filename=None,
                 source_classification_filename=None, p_astro_filename=None):
    """Produce an update alert.

    This does nothing more than call
    :meth:`~gwcelery.tasks.orchestrator.initial_or_update_alert` with
    ``alert_type='update'``.

    Parameters
    ----------
    superevent_id : str
        The superevent ID.
    skymap_filename : str, optional
        The sky map to send.
        If None, then most recent public sky map is used.
    source_classification_filename : str, optional
        The source classification file to use.
        If None, then most recent one is used.
    p_astro_filename : str, optional
        The p_astro file to use.
        If None, then most recent one is used.
    """
    initial_or_update_alert(superevent_id, 'update', skymap_filename,
                            source_classification_filename, p_astro_filename)


@app.task(ignore_result=True, shared=False)
def retraction_alert(superevent_id):
    """Produce a retraction alert. This is currently just a stub and does
    nothing more than create and send a VOEvent."""
    (
        gracedb.expose.s(superevent_id)
        |
        _create_voevent.si(
            None, superevent_id, 'retraction',
            internal=False,
            open_alert=True,
            vetted=True
        )
        |
        group(
            gracedb.download.s(superevent_id)
            |
            gcn.send.s(),

            gracedb.create_tag.s('public', superevent_id)
        )
    ).apply_async()



"""Validate LIGO/Virgo GCN notices to make sure that their contents match the
original VOEvent notices that we sent."""
from urllib.parse import urlparse

import lxml.etree

from .. import gcn
from .. import gracedb


@gcn.handler(gcn.NoticeType.LVC_PRELIMINARY,
             gcn.NoticeType.LVC_INITIAL,
             gcn.NoticeType.LVC_UPDATE,
             shared=False)
def validate_voevent(payload):
    """Check that the contents of a public LIGO/Virgo GCN matches the original
    VOEvent in GraceDB."""
    if not __debug__:  # pragma: no cover
        raise RuntimeError('This task will not function correctly because '
                           'Python assertions are disabled.')

    root = lxml.etree.fromstring(payload)

    # Which GraceDB ID does this refer to?
    graceid = root.find("./What/Param[@name='GraceID']").attrib['value']

    # Skip static events from the user guide, because they are not in GraceDb.
    if graceid == 'MS181101ab':
        return

    # Which VOEvent does this refer to?
    u = urlparse(root.attrib['ivorn'])
    local_id = u.fragment
    filename = local_id + '.xml'

    # Download and parse original VOEvent
    orig = gracedb.download(filename, graceid)

    assert orig == payload, 'GCN does not match GraceDb'

    # Tag the VOEvent to indicate that it was received correctly
    gracedb.create_tag(filename, 'gcn_received', graceid)



"""Qualitative source classification for CBC events."""
import io
import json

from ligo.skymap.io import events

from ..import app


def source_classification(m1, m2, c1, threshold=3.0):
    """This is the place-holder function for the source classfication pipeline.
    In the future, the actual source classification pipeline will be integrated
    in three steps. First step will be the simple integration of the
    point-estimate code that will be using the em_progenitors code from PyCBC.
    In the second step, rapid_pe needs to be made Python3 compatible so that
    the ambiguity ellipsoid feature can be brough back into action. And,
    finally the O3 implementation will be incorporated which is currently a
    work in progress. This placeholder code will only act upon the mass2 point
    estimate value and classify the systems as whether they have a neutron or
    not. It does not attempt to classify for the remnant mass, returns a NaN
    value for that probability."""
    if m2 <= threshold:
        return([100.0, 100.0])
    else:
        return([0.0, 0.0])


@app.task(shared=False)
def classifier(coinc_psd, graceid):
    """This function is currently actually calculating the simple
       source classification probability (m1 < 3.0 M_sun). In the
       future this code will call a classification code that will
       be put on lalinference.
    """
    # Parse event
    coinc, psd = coinc_psd
    coinc = io.BytesIO(coinc)
    psd = io.BytesIO(psd)
    event_source = events.ligolw.open(coinc, psd_file=psd, coinc_def=None)
    event, = event_source.values()

    # Run EM_Bright
    template_args = event.template_args
    mass1 = template_args['mass1']  # primary object mass
    mass2 = template_args['mass2']  # secondary object mass
    chi1 = template_args['spin1z']  # primary object aligned spin
    [p_ns, p_em] = source_classification(mass1, mass2, chi1)

    data = {
        'Prob NS2': p_ns,
        'Prob EMbright': p_em
    }

    return json.dumps(data)



"""Provide subcommand to manage GWCelery through HTCondor."""
import json
import os
import shlex
import subprocess
import sys
import time

from celery.bin.base import Command
import lxml.etree
import pkg_resources

SUBMIT_FILE = pkg_resources.resource_filename(__name__, 'data/gwcelery.sub')


def get_constraints():
    return '-constraint', 'JobBatchName=={} && Iwd=={}'.format(
        json.dumps('gwcelery'),  # JSON string literal escape sequences
        json.dumps(os.getcwd())  # are a close match to HTCondor ClassAds.
    )


def run_exec(*args):
    print(' '.join(shlex.quote(arg) for arg in args))
    os.execvp(args[0], args)


def running():
    """Determine if GWCelery is already running under HTCondor."""
    status = subprocess.check_output(('condor_q', '-xml', *get_constraints()))
    classads = lxml.etree.fromstring(status)
    return classads.find('.//c') is not None


def submit():
    """Submit all GWCelery jobs to HTCondor (if not already running)."""
    if running():
        print('error: GWCelery jobs are already running in this directory.\n'
              'You must first remove exist jobs with "gwcelery condor rm".\n'
              'To see the status of those jobs, run "gwcelery condor q".',
              file=sys.stderr)
        sys.exit(1)
    else:
        run_exec('condor_submit', SUBMIT_FILE)


def resubmit():
    """Remove any running GWCelery jobs and resubmit to HTCondor."""
    if running():
        subprocess.check_call(('condor_rm', *get_constraints()))
    timeout = 60
    start = time.monotonic()
    while time.monotonic() - start < timeout:
        if not running():
            break
        time.sleep(1)
    else:
        print('error: Could not stop all GWCelery jobs', file=sys.stderr)
        sys.exit(1)
    run_exec('condor_submit', SUBMIT_FILE)


def rm():
    """Remove all GWCelery jobs."""
    run_exec('condor_rm', *get_constraints())


def hold():
    """Put all GWCelery jobs on hold."""
    run_exec('condor_hold', *get_constraints())


def release():
    """Release all GWCelery jobs from hold status."""
    run_exec('condor_release', *get_constraints())


def q():
    """Show status of all GWCelery jobs."""
    run_exec('condor_q', '-nobatch', *get_constraints())


class CondorCommand(Command):
    """Shortcuts for HTCondor commands to manage deployment of GWCelery on LIGO
    Data Grid clusters. These commands apply to the GWCelery instance that is
    running in the current working directory."""

    def add_arguments(self, parser):
        subparsers = parser.add_subparsers()
        for func in [submit, rm, hold, release, resubmit, q]:
            subparser = subparsers.add_parser(func.__name__, help=func.__doc__)
            subparser.set_defaults(func=func)

    def run(self, func=None, **kwargs):
        func()



"""Flask web application for manually triggering certain tasks."""
import argparse
import datetime
import os
import re

from astropy.time import Time
from celery.bin.base import Command
import click
import click.testing
from flask import (flash, Flask, jsonify, redirect, render_template, request,
                   url_for)
from flask_caching import Cache
from flask.cli import FlaskGroup
from ligo.gracedb.rest import HTTPError as GraceDbHTTPError

from . import app as celery_app
from .tasks import first2years, gracedb, orchestrator


# from http://flask.pocoo.org/snippets/35/
class ReverseProxied(object):
    '''Wrap the application in this middleware and configure the
    front-end server to add these headers, to let you quietly bind
    this to a URL other than / and to an HTTP scheme that is
    different than what is used locally.

    In nginx:
    location /myprefix {
        proxy_pass http://192.168.0.1:5001;
        proxy_set_header Host $host;
        proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
        proxy_set_header X-Scheme $scheme;
        proxy_set_header X-Script-Name /myprefix;
        }

    :param app: the WSGI application
    '''
    def __init__(self, app, script_name=None):
        self.app = app
        self.script_name = script_name

    def __call__(self, environ, start_response):
        script_name = environ.get('HTTP_X_SCRIPT_NAME', '') or self.script_name
        if script_name:
            environ['SCRIPT_NAME'] = script_name
            path_info = environ['PATH_INFO']
            if path_info.startswith(script_name):
                environ['PATH_INFO'] = path_info[len(script_name):]

        scheme = environ.get('HTTP_X_SCHEME', '')
        if scheme:
            environ['wsgi.url_scheme'] = scheme
        return self.app(environ, start_response)


app = Flask(__name__)
app.wsgi_app = ReverseProxied(app.wsgi_app)
app.wsgi_app.script_name = os.environ.get('FLASK_URL_PREFIX')

# Default secret key: secure and random. However, sessions are not preserved
# across different Python processes.
app.config['SECRET_KEY'] = os.urandom(24)

# Set up a server-side cache to store autocomplete responses in order to reduce
# traffic to GraceDb. The cache's backend is the same Redis database that
# Celery uses, although the Redis keys will have a different prefix so that
# they are ignored by Celery.
cache = Cache(app, config={'CACHE_DEFAULT_TIMEOUT': 30,  # lifetime in seconds
                           'CACHE_REDIS_HOST': celery_app.backend.client,
                           'CACHE_TYPE': 'redis'})


@app.route('/')
def index():
    return render_template('index.html')


def take_n(n, iterable):
    for i, item in enumerate(iterable):
        if i >= n:
            break
        yield item


# Regular expression for parsing query strings
# that look like GraceDb superevent names.
typeahead_graceid_regex = re.compile(
    r'(?P<prefix>[MT]?)S?(?P<date>\d{0,6})(?P<suffix>[a-z]*)',
    re.IGNORECASE)


@app.route('/typeahead_superevent_id')
@cache.cached(query_string=True)
def typeahead_superevent_id():
    """Search GraceDb for superevents by ID.

    This involves some date parsing because GraceDb does not support directly
    searching for superevents by ID substring."""

    max_results = 8  # maximum number of results to return
    batch_results = 32  # batch size for results from server

    term = request.args.get('superevent_id')
    match = typeahead_graceid_regex.fullmatch(term) if term else None

    if match:
        # Determine GraceDb event category from regular expression.
        prefix = match['prefix'].upper() + 'S'
        category = {'T': 'test', 'M': 'MDC'}.get(
            match['prefix'].upper(), 'production')

        # Determine start date from regular expression by padding out
        # the partial date with missing digits defaulting to 000101.
        date_partial = match['date']
        date_partial_length = len(date_partial)
        try:
            date_start = datetime.datetime.strptime(
                date_partial + '000101'[date_partial_length:], '%y%m%d')
        except ValueError:  # invalid date
            return jsonify([])

        # Determine end date from regular expression by adding a very
        # loose upper bound on the number of days until the next
        # digit in the date rolls over. No need to be exact here.
        date_end = date_start + datetime.timedelta(
            days=[36600, 3660, 366, 320, 32, 11, 1.1][date_partial_length])

        # Determine GraceDb event suffix from regular expression.
        suffix = match['suffix'].lower()
    else:
        prefix = 'S'
        category = 'production'
        date_end = datetime.datetime.utcnow()
        date_start = date_end - datetime.timedelta(days=7)
        date_partial = ''
        date_partial_length = 0
        suffix = ''

    # Query GraceDb.
    query = 'category: {} t_0: {} .. {}'.format(
        category, Time(date_start).gps, Time(date_end).gps)
    response = gracedb.client.superevents(
        query, orderby='superevent_id', count=batch_results)

    # Filter superevent IDs that match the search term.
    regex = re.compile(r'{}{}\d{{{}}}{}[a-z]*'.format(
        prefix, date_partial, 6 - date_partial_length, suffix))
    superevent_ids = (
        superevent['superevent_id'] for superevent
        in response if regex.fullmatch(superevent['superevent_id']))

    # Return only the first few matches.
    return jsonify(list(take_n(max_results, superevent_ids)))


def _search_by_tag_and_filename(superevent_id, filename, extension, tag):
    try:
        records = gracedb.client.logs(superevent_id).json()['log']
        return [
            record['filename'] for record in records
            if tag in record['tag_names']
            and record['filename'].startswith(filename)
            and record['filename'].endswith(extension)]
    except GraceDbHTTPError as e:
        # Ignore 404 errors from server
        if e.status == 404:
            return []
        else:
            raise


@app.route('/typeahead_skymap_filename')
@cache.cached(query_string=True)
def typeahead_skymap_filename():
    return jsonify(_search_by_tag_and_filename(
        request.args.get('superevent_id') or '',
        request.args.get('filename') or '',
        '.fits.gz', 'sky_loc'
    ))


@app.route('/typeahead_source_classification_filename')
@cache.cached(query_string=True)
def typeahead_source_classification_filename():
    return jsonify(_search_by_tag_and_filename(
        request.args.get('superevent_id') or '',
        request.args.get('filename') or '',
        '.json', 'em_bright'
    ))


@app.route('/typeahead_p_astro_filename')
@cache.cached(query_string=True)
def typeahead_p_astro_filename():
    return jsonify(_search_by_tag_and_filename(
        request.args.get('superevent_id') or '',
        request.args.get('filename') or '',
        '.json', 'p_astro'
    ))


@app.route('/send_update_gcn', methods=['POST'])
def send_update_gcn():
    keys = ('superevent_id', 'skymap_filename',
            'source_classification_filename', 'p_astro_filename')
    superevent_id, *_ = args = tuple(request.form.get(key) for key in keys)
    if all(args):
        app.logger.info(
            'Calling update_alert%r', args)
        orchestrator.update_alert(*args)
        flash('Queued update alert for {}.'.format(superevent_id), 'success')
    else:
        flash('No alert sent. Please fill in all fields.', 'danger')
    return redirect(url_for('index'))


@app.route('/send_mock_event', methods=['POST'])
def send_mock_event():
    first2years.upload_event.delay()
    flash('Queued a mock event.', 'success')
    return redirect(url_for('index'))


@click.group(cls=FlaskGroup, create_app=lambda *args, **kwargs: app)
def main():
    pass


class FlaskCommand(Command):

    def add_arguments(self, parser):
        # Capture command line help from Flask
        runner = click.testing.CliRunner()
        result = runner.invoke(main, ['--help'])
        flask_help = result.output.replace('main', 'gwcelery flask')

        group = parser.add_argument_group(
            'Flask Options', description=flask_help)
        group.add_argument(
            'flask_args', nargs=argparse.REMAINDER, help=argparse.SUPPRESS)

    def run(self, *args, flask_args=(), **kwargs):
        # Allow port number to be specified from an environment variable.
        port = os.environ.get('FLASK_PORT')
        if port:
            flask_args += ['--port', port]
        main(flask_args)


main.__doc__ = FlaskCommand.__doc__ = __doc__



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# GWCelery documentation build configuration file, created by
# sphinx-quickstart on Tue Dec  5 14:59:24 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import configparser
import importlib
import os
import pkg_resources
import sys

sys.path.insert(0, os.path.abspath('..'))


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'celery.contrib.sphinx',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosectionlabel',
    'sphinx.ext.graphviz',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = ['.rst']
source_parsers = {}

# The master toctree document.
master_doc = 'index'

# General information about the project.
setup_cfg = configparser.ConfigParser()
setup_cfg.read('../setup.cfg')
project = setup_cfg['metadata']['name']
author = setup_cfg['metadata']['author']
description = setup_cfg['metadata']['description']

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.

spec = importlib.util.spec_from_file_location(
    '_version', '../gwcelery/_version.py')
module = importlib.util.module_from_spec(spec)
spec.loader.exec_module(module)
parsed_version = pkg_resources.parse_version(module.get_versions()['version'])
del module, spec

version = parsed_version.base_version
# The full version, including alpha/beta/rc tags.
release = parsed_version.public

del setup_cfg, parsed_version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

autodoc_mock_imports = []
for mod in ['astropy',
            'comet',
            'gcn',
            'glue',
            'gwdatafind',
            'gwpy',
            'healpy',
            'lal',
            'lalapps',
            'lalburst',
            'laldetchar',
            'lalframe',
            'lalinference',
            'lalinspiral',
            'lalmetaio',
            'lalpulsar',
            'lalsimulation',
            'lalstochastic',
            'lalxml',
            'ligo',
            'ligo.followup_advocate',
            'ligo.p_astro'
            'ligo.raven',
            'ligo.segments',
            'ligo.skymap',
            'lxml',
            'lxml.etree',
            'matplotlib',
            'numpy',
            'safe_netrc',
            'scipy',
            'sentry_sdk',
            'sleek_lvalert',
            'twisted',
            'pytest',
            'zope']:
    try:
        importlib.import_module(mod)
    except ImportError:
        autodoc_mock_imports.append(mod)
autodoc_default_options = {
    'members': None,
    'undoc-members': None,
    'show-inheritance': None
}
autodoc_member_order = 'bysource'
autosummary_generate = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    'description': description,
    'github_button': False,
    'logo': 'logo.png',
    'logo_name': True
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    'index': [
        'about.html',
        'searchbox.html',
    ],
    '**': [
        'about.html',
        'navigation.html',
        'searchbox.html',
    ]
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = project + 'doc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, project + '.tex', 'GWCelery Documentation',
     author, 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, project.lower(), project + ' Documentation',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, project, project + ' Documentation',
     author, project, 'One line description of project.',
     'Miscellaneous'),
]


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'celery': ('https://celery.readthedocs.io/en/latest/', None),
    'celery_eternal': ('https://celery-eternal.readthedocs.io/en/latest/',
                       None),
    'comet': ('https://comet.readthedocs.io/en/stable/', None),
    'data-quality-report': (
        'https://docs.ligo.org/detchar/data-quality-report/', None),
    'gwpy': ('https://gwpy.github.io/docs/stable/', None),
    'ligo.skymap': ('https://leo-singer.docs.ligo.org/ligo.skymap/', None),
    'userguide': ('https://emfollow.docs.ligo.org/userguide/', None)
}


# -- Options for viewcode extension ---------------------------------------

# celery.contrib.sphinx does not show source links for tasks if this is set
# to the default value of True
viewcode_import = False


# -- Options for graphviz extension ---------------------------------------

graphviz_output_format = 'svg'


# -- Options for autosectionlabel extension -------------------------------

autosectionlabel_prefix_document = True



import json

import pytest

from ..tasks import p_astro_other


@pytest.mark.parametrize(
    'snr,far,mass1,mass2', ([30., 1e-17, 1.4, 1.4],
                            [100, 1e-10, 400, 40]))
def test_compute_p_astro(snr, far, mass1, mass2):
    """Check if p_astros sum up to unity"""
    p_astros = json.loads(
        p_astro_other.compute_p_astro(snr, far, mass1, mass2))
    assert pytest.approx(sum(p_astros.values())) == 1.


def test_compute_p_astro_bns():
    """Test p_astro values with CBC catalog paper
    values for GW170817
    """
    # values based on G322759
    snr = 33.
    far = 7.6e-59
    mass1 = 1.77
    mass2 = 1.07

    p_astros = json.loads(
        p_astro_other.compute_p_astro(snr, far, mass1, mass2))
    assert pytest.approx(p_astros['BNS']) == 1.



"""Annotations for sky maps."""
import os
import tempfile

from astropy.io import fits
from astropy import table
from celery import group
from ligo.skymap.tool import ligo_skymap_flatten
from ligo.skymap.tool import ligo_skymap_plot
from ligo.skymap.tool import ligo_skymap_plot_volume
from matplotlib import pyplot as plt

from . import gracedb
from ..import app
from ..jinja import env
from ..util.tempfile import NamedTemporaryFile


def annotate_fits(versioned_filename, graceid, tags):
    """Perform annotations on a sky map.

    This function downloads a FITS file and then generates and uploads all
    derived images as well as an HTML dump of the FITS header.
    """
    filebase = versioned_filename.partition('.fits')[0]
    header_msg = (
        'FITS headers for <a href="/api/superevents/{graceid}/files/'
        '{versioned_filename}">{versioned_filename}</a>').format(
            graceid=graceid, versioned_filename=versioned_filename)
    allsky_msg = (
        'Mollweide projection of <a href="/api/superevents/{graceid}/files/'
        '{versioned_filename}">{versioned_filename}</a>').format(
            graceid=graceid, versioned_filename=versioned_filename)
    volume_msg = (
        'Volume rendering of <a href="/api/superevents/{graceid}/files/'
        '{versioned_filename}">{versioned_filename}</a>').format(
            graceid=graceid, versioned_filename=versioned_filename)

    return group(
        fits_header.s(versioned_filename) |
        gracedb.upload.s(
            filebase + '.html', graceid, header_msg, tags),

        plot_allsky.s() |
        gracedb.upload.s(
            filebase + '.png', graceid, allsky_msg, tags),

        annotate_fits_volume.s(
            filebase + '.volume.png', graceid, volume_msg, tags)
    )


def is_3d_fits_file(filecontents):
    """Determine if a FITS file has distance information."""
    with NamedTemporaryFile(content=filecontents) as fitsfile:
        return 'DISTNORM' in table.Table.read(fitsfile.name).colnames


@app.task(ignore_result=True, shared=False)
def annotate_fits_volume(filecontents, *args):
    """Perform annotations that are specific to 3D sky maps."""
    if is_3d_fits_file(filecontents):
        (
            plot_volume.s(filecontents)
            |
            gracedb.upload.s(*args)
        ).apply_async()


@app.task(shared=False)
def fits_header(filecontents, filename):
    """Dump FITS header to HTML."""
    template = env.get_template('fits_header.html')
    with NamedTemporaryFile(content=filecontents) as fitsfile, \
            fits.open(fitsfile.name) as hdus:
        return template.render(filename=filename, hdus=hdus)


@app.task(shared=False)
def plot_allsky(filecontents):
    """Plot a Mollweide projection of a sky map using the command-line tool
    :doc:`ligo-skymap-plot <ligo/skymap/tool/ligo_skymap_plot>`."""
    # Note: plt.style.context added as workaround for
    # https://github.com/astropy/astropy/issues/8004.
    with NamedTemporaryFile(mode='rb', suffix='.png') as pngfile, \
            NamedTemporaryFile(content=filecontents) as fitsfile, \
            plt.style.context({'text.usetex': False}, after_reset=True):
        ligo_skymap_plot.main([fitsfile.name, '-o', pngfile.name,
                               '--annotate', '--contour', '50', '90'])
        return pngfile.read()


@app.task(queue='openmp', shared=False)
def plot_volume(filecontents):
    """Plot a 3D volume rendering of a sky map using the command-line tool
    :doc:`ligo-skymap-plot-volume <ligo/skymap/tool/ligo_skymap_plot_volume>`.
    """
    # Note: plt.style.context added as workaround for
    # https://github.com/astropy/astropy/issues/8004.
    with NamedTemporaryFile(mode='rb', suffix='.png') as pngfile, \
            NamedTemporaryFile(content=filecontents) as fitsfile, \
            plt.style.context({'text.usetex': False}, after_reset=True):
        ligo_skymap_plot_volume.main([fitsfile.name, '-o',
                                      pngfile.name, '--annotate'])
        return pngfile.read()


@app.task(shared=False)
def flatten(filecontents, filename):
    """Convert a HEALPix FITS file from multi-resolution UNIQ indexing to the
    more common IMPLICIT indexing using the command-line tool
    :doc:`ligo-skymap-flatten <ligo/skymap/tool/ligo_skymap_flatten>`."""
    with NamedTemporaryFile(content=filecontents) as infile, \
            tempfile.TemporaryDirectory() as tmpdir:
        outfilename = os.path.join(tmpdir, filename)
        ligo_skymap_flatten.main([infile.name, outfilename])
        return open(outfilename, 'rb').read()



import json

import pkg_resources
import pytest

from ..tasks.gcn.validate import validate_voevent


@pytest.fixture
def fake_gcn(celeryconf, monkeypatch):

    def mock_download(filename, graceid):
        assert filename == 'G298048-1-Initial.xml'
        assert graceid == 'G298048'
        return pkg_resources.resource_string(__name__, 'data/' + filename)

    def mock_get_log(graceid):
        assert graceid == 'G298048'
        return json.loads(
            pkg_resources.resource_string(__name__, 'data/G298048_log.json'))

    def mock_create_tag(filename, tag, graceid):
        assert filename == 'G298048-1-Initial.xml'
        assert tag == 'gcn_received'
        assert graceid == 'G298048'

    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.download', mock_download)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.get_log', mock_get_log)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.create_tag', mock_create_tag)

    # Get the VOEvent.
    yield pkg_resources.resource_string(
        __name__, 'data/G298048-1-Initial.xml')


def test_validate_voevent(fake_gcn):
    """Test that the fake GCN notice matches what we actually sent."""
    validate_voevent(fake_gcn)


def test_validate_voevent_mismatched_param(fake_gcn):
    """Test that we correctly detect mismatched parameter values in GCNs."""
    with pytest.raises(AssertionError, match="^GCN does not match GraceDb$"):
        validate_voevent(fake_gcn + b'\n')
